#include "winresrc.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define FILE 256
#define IDI_ICON  1001

IDI_ICON               ICON    DISCARDABLE     "gui/media/scummvmtools.ico"

VS_VERSION_INFO VERSIONINFO
 FILEVERSION     2,7,0,0
 PRODUCTVERSION  2,7,0,0
 FILEFLAGSMASK   VS_FFI_FILEFLAGSMASK
#ifdef _DEBUG
 FILEFLAGS       VS_FF_DEBUG
#else
 FILEFLAGS       0
#endif
 FILEOS          VOS_NT_WINDOWS32
 FILETYPE        VFT_APP
 FILESUBTYPE     VFT2_UNKNOWN
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0" // US English, Unicode
        BEGIN
            VALUE "Comments", "Look! A three headed monkey (TM)! .. Nice use of the TM!\0"
            VALUE "FileDescription", "https://www.scummvm.org/\0"
            VALUE "FileVersion", "2.7.0\0"
            VALUE "InternalName", "scummvm\0"
            VALUE "LegalCopyright", "Copyright © 2001-2023 The ScummVM Team\0"
            VALUE "LegalTrademarks", "'SCUMM', and all SCUMM games are a TM of LucasArts. Simon The Sorcerer is a TM of AdventureSoft. Beneath a Steel Sky and Broken Sword are a TM of Revolution. Flight of the Amazon Queen is a TM of John Passfield and Steve Stamatiadis. \0"
            VALUE "OriginalFilename", "scummvm-tools.exe\0"
            VALUE "ProductName", "ScummVM Tools\0"
            VALUE "ProductVersion", "2.7.0\0"
        END
    END

    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200 // US English, Unicode
    END
END
